/* -*- linux-c -*-
 *
 *	MOXA C-218 Serial Board Driver for Linux, version 0.98
 *
 *	(c) 1995 -- 1999 Martin Mares, <mj@ucw.cz>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	This driver is based on cyclades.c by Marcio Saito (cyclades@netcom.com)
 *	and Randolph Bentson (bentson@grieg.seaslug.org). Much of the design and
 *	the code came from serial.c which is copyrighted by Linus Torvalds,
 *	extensively rewritten by Theodore Ts'o and fixed by Michael K. Johnson.
 *
 */

/* ### TODO ###
 *
 *	* Baud rates
 *	- break sending and reception
 *	- better ring buffer pointer operations
 *	- buffer flushing
 */

/* ### HISTORY ###
 *	0.91	14-03-96	Changed to work with 1.3.70
 *	0.92	02-04-96	Modem control handling fixed
 *	0.93	12-04-96	Modem event scheduling rewritten and fixed
 *	0.94	17-09-96	Usecount cleanup
 *	0.95	01-11-96	Usecount cleanup #2
 *	0.96	12-04-97	Ported to 2.1.X
 *				Reindented to standard kernel style
 *				Major code cleanup
 *				Now uses phys_to_virt instead of ioremap or other clean
 *					I/O memory access, but it doesn't matter since
 *					the driver cannot work on non-i386 architectures
 *					because of endianity dependencies and other caveats.
 *				Usecount bugs--
 *	0.97	08-05-97	Rewritten bottom half processing. Now uses tq_immediate.
 *				Now needs no kernel patches.
 *	0.98	27-09-99	Ported to 2.2.12.
 */

/*
 *	Debugging switches.
 */

#if 0
#define MOXA_DEBUG
#define MOXA_DEBUG_COMMANDS
#define MOXA_DEBUG_EVENTS
#define MOXA_DEBUG_OVERFLOW
#undef MOXA_DEBUG_TIMER
#endif
#undef MOXA_LOOPBACK_ONLY
#undef MOXA_DEBUGGING_IOCTL
#define SERIAL_PARANOIA_CHECK
 
/*
 *	Kernel includes.
 */

#include <linux/config.h>
#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#endif

#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/interrupt.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <linux/mm.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>

#include "moxa.h"

/*
 *	Major numbers.
 *	If they are not found, use the `local & experimental' range.
 */

#include <linux/major.h>

#ifndef MOXA_MAJOR
#define MOXA_MAJOR 60
#define MOXAAUX_MAJOR 61
#endif

/*
 *	Some debugging stuff...
 */

#ifdef MOXA_DEBUG
#define DBPRINT(x) printk x;
#else
#define DBPRINT(x)
#endif

/*
 *	Useful macros and global variables.
 */

#define small_delay(x) for(j=0;j<x;j++)k++;

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

#define SERIAL_TYPE_NORMAL  1
#define SERIAL_TYPE_CALLOUT 2

static struct tty_driver mx_serial_driver, mx_callout_driver;

static struct timer_list mx_timer;

static int num_channels;

static int serial_refcount;

/*
 *	Table of addresses for all MOXA C-218 cards in the system.
 *	No autoprobing is used. You must manually modify this array to let
 *	the driver know about what is where.
 */

static int moxa_addr[4] =
{
   0xd0000,
	0,
	0,
	0
};

/*
 *	IRQ used by all the cards (_only_ shared IRQs are supported)
 */

#define MOXA_IRQ 12

/*
 *	The mx_events array contains pointers to event reporting structures
 * 	of all cards.
 */

static struct moxa_event_ctrl *mx_events[4];

/*
 *	The mx_port array contains a moxa_port structure for each port.
 */

static struct moxa_port mx_port[] = {
	{-1 },      /* ttyM0 */
	{-1 },      /* ttyM1 */
	{-1 },      /* ttyM2 */
	{-1 },      /* ttyM3 */
	{-1 },      /* ttyM4 */
	{-1 },      /* ttyM5 */
	{-1 },      /* ttyM6 */
	{-1 },      /* ttyM7 */
	{-1 },      /* ttyM8 */
	{-1 },      /* ttyM9 */
	{-1 },      /* ttyM10 */
	{-1 },      /* ttyM11 */
	{-1 },      /* ttyM12 */
	{-1 },      /* ttyM13 */
	{-1 },      /* ttyM14 */
	{-1 },      /* ttyM15 */
#if 0
	{-1 },      /* ttyM16 */
	{-1 },      /* ttyM17 */
	{-1 },      /* ttyM18 */
	{-1 },      /* ttyM19 */
	{-1 },      /* ttyM20 */
	{-1 },      /* ttyM21 */
	{-1 },      /* ttyM22 */
	{-1 },      /* ttyM23 */
	{-1 },      /* ttyM24 */
	{-1 },      /* ttyM25 */
	{-1 },      /* ttyM26 */
	{-1 },      /* ttyM27 */
	{-1 },      /* ttyM28 */
	{-1 },      /* ttyM29 */
	{-1 },      /* ttyM30 */
	{-1 }       /* ttyM31 */
#endif
};
#define NR_PORTS        (sizeof(mx_port)/sizeof(struct moxa_port))

static struct tty_struct *serial_table[NR_PORTS];
static struct termios *serial_termios[NR_PORTS];
static struct termios *serial_termios_locked[NR_PORTS];

/*
 *	Table of baud rates.
 */

static short baud_table[] = {
-1, 0, 1, 2, 3, 4, -1, 5, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16 };

/*
 *	Declarations
 */

static void config_setup(struct moxa_port * info);
static void get_modem_status(struct moxa_port *info);
static void modem_event(struct moxa_port *info);

/*
 *	tmp_buf is used as a temporary buffer by serial_write.  We need to
 * 	lock it in case user space access blocks while swapping in a page,
 * 	and some other program tries to do a serial write at the same time.
 * 	Since the lock will only come under contention when the system is
 * 	swapping and available memory is low, it makes sense to share one
 * 	buffer across all the serial ports, since it significantly saves
 * 	memory if large numbers of serial ports are open.
 */

static byte *tmp_buf = 0;
static struct semaphore tmp_buf_sem = MUTEX;

/*
 *	If the SERIAL_PARANOIA_CHECK switch is active, we check whether the port
 *	structure we want to use contains the magic value...
 */

#ifdef SERIAL_PARANOIA_CHECK

static inline int
serial_paranoia_check(struct moxa_port *info,
		      dev_t device, const char *routine)
{
	static const char *badmagic =
		"Warning: bad magic number for MOXA struct (%d, %d) in %s\n";
	static const char *badinfo =
		"Warning: null moxa_port for (%d, %d) in %s\n";

	if (!info) {
		printk(badinfo, MAJOR(device), MINOR(device), routine);
		return 1;
	}

	if (info->magic != MOXA_MAGIC) {
		printk(badmagic, MAJOR(device), MINOR(device), routine);
		return 1;
	}
	return 0;
}

#else

#define serial_paranoia_check(i,d,r)

#endif

/*
 *	When debugging is switched on, we need to spew information on the screen
 *	even during interrupts. These functions allow it.
 */

#ifdef MOXA_DEBUG

#if 0
static void
SP(char *data)
{
	ulg flags;

	save_flags(flags); cli();
	printk(data);
	restore_flags(flags);
}
#endif

static char scrn[2] = { 0, 0 };

static void
CP(char data)
{
	ulg flags;

	save_flags(flags); cli();
	scrn[0] = data;
	printk(scrn);
	restore_flags(flags);
}

static void CP1(int data) { (data<10)?  CP(data+'0'): CP(data+'A'-10); }
static void CP2(int data) { CP1((data>>4) & 0x0f); CP1( data & 0x0f); }
static void CP4(int data) { CP2((data>>8) & 0xff); CP2(data & 0xff); }
static void CP8(long data) { CP4((data>>16) & 0xffff); CP4(data & 0xffff); }

#endif

/*
 *	Send command to the card.
 */

#define MOXA_CC (info->ctrl)

static inline void
mx_cmd_start(struct moxa_port *info)
{
	ulg flags;

	save_flags(flags); cli();
	if (info->cmd_in_progress) {
		restore_flags(flags);
		printk(KERN_ERR "moxa: attempted to do nested command processing !!!\n");
	} else {
		info->cmd_in_progress = 1;
		restore_flags(flags);
	}
}

static inline void
mx_cmd_stop(struct moxa_port *info)
{
	ulg flags;

	save_flags(flags); cli();
	info->cmd_in_progress = 0;
	if (info->modem_event_scheduled == 1) {
		info->modem_event_scheduled = 2;
		restore_flags(flags);
		get_modem_status(info);
	} else
		restore_flags(flags);
}

static inline int
mx_cmd_wait(struct moxa_port *info)
{
	int timeout = jiffies + 1*HZ;
	int j,k;

	while (jiffies < timeout) {
		if (MOXA_CC -> command == 0) return 0;
		small_delay(30);	/* We shouldn't hog the shared memory bus */
	}
	printk("moxa: timeout while executing command.\n");
	return 0;
}

static void
mx_cmd(struct moxa_port *info,word cmd)
{
	mx_cmd_start(info);
	mx_cmd_wait(info);
#ifdef MOXA_DEBUG_COMMANDS
	DBPRINT((">%02x ",cmd));
#endif
	MOXA_CC -> command = cmd;
	mx_cmd_stop(info);
}

static word
mx_cmd_r(struct moxa_port *info,word cmd)
{
	word value;

	mx_cmd_start(info);
	mx_cmd_wait(info);
#ifdef MOXA_DEBUG_COMMANDS
	DBPRINT((">%02x ",cmd));
#endif
	MOXA_CC -> command = cmd;
	if (mx_cmd_wait(info)) {
		mx_cmd_stop(info);
		return 0;
	}
	value = MOXA_CC -> par1;
	mx_cmd_stop(info);
#ifdef MOXA_DEBUG_COMMANDS
	DBPRINT(("<%04x ",value));
#endif
	return value;
}

static void
mx_cmd1(struct moxa_port *info,word cmd,word p1)
{
	mx_cmd_start(info);
	mx_cmd_wait(info);
#ifdef MOXA_DEBUG_COMMANDS
	DBPRINT((">%02x/%04x ",cmd,p1));
#endif
	MOXA_CC -> par1 = p1;
	MOXA_CC -> command = cmd;
	mx_cmd_stop(info);
}

/*
 *	Get one byte from MOXA card memory, including appropriate paging.
 *	Must be called with interrupts disabled.
 */

static inline byte
mx_get_byte(struct moxa_port *info, ulg address)
{
	byte *addr;

	*info->page_port = (address >> 13);
	addr = info->mem_window + (address & 0x1fff);
	return *addr;
}

/*
 *	Put one byte to MOXA card memory, including appropriate paging.
 *	Must be called with interrupts disabled.
 */

static inline void
mx_put_byte(struct moxa_port *info, ulg address, byte data)
{
	byte *addr;

	*info->page_port = (address >> 13);
	addr = info->mem_window + (address & 0x1fff);
	*addr = data;
}

/*
 *	Read modem status. Delayed modem events are processed here.
 */

static void
get_modem_status(struct moxa_port *info)
{
	ulg flags;

	info->modem_status = mx_cmd_r(info, MOXA_CMD_GET_MODEM);
	save_flags(flags); cli();
	if (info->modem_event_scheduled) {
		info->modem_event_scheduled = 0;
		restore_flags(flags);
		modem_event(info);
	} else
		restore_flags(flags);
	info->modem_status_old = info->modem_status;
}

/*
 *		Schedule modem status examination. Called by the bottom half of
 *		our interrupt handler. Requires special processing, because sending
 *		of MOXA command is required, but it may already be in progress.
 */

static inline void
schedule_modem_event(struct moxa_port *info)
{
	ulg flags;

	save_flags(flags); cli();
	if (info->modem_event_scheduled == 2) {
		restore_flags(flags);
#ifdef MOXA_DEBUG_EVENTS
		CP('q');
#endif
		return;
	}
	info->modem_event_scheduled = 1;
	if (info->cmd_in_progress) {
		restore_flags(flags);		/* To be continued later */
#ifdef MOXA_DEBUG_EVENTS
		CP('Q');
#endif
	} else {
		restore_flags(flags);
		get_modem_status(info);
	}
}

/*
 *	This routine is used by the interrupt handlers to schedule
 *	processing in the software interrupt portion of the driver
 *	(also known as the "bottom half").  This can be called any
 *	number of times for any channel without harm.
 */

static inline void
mx_sched_event(struct moxa_port *info, word event)
{
#ifdef MOXA_DEBUG_EVENTS
	CP('!');
#endif
	info->event |= event;				    /* remember what kind of event and who it belong to */
	queue_task(&info->tqueue, &tq_immediate);	    /* Use the standard immediate-type BH */
	mark_bh(IMMEDIATE_BH);
}

/*
 *	Timer interrupt. Called twice in a second to check if there are
 *	any free transmitter buffers for pending writes. It's called by
 *	the bottom half of standard timer int handler, but may call our
 *	own bottom half, because it has lower priority than the timer.
 */

static void
mx_timer_int(unsigned long data)
{
	int chan;
	ulg flags;
	struct moxa_port *info;
	volatile struct moxa_chan_ctrl *ctrl;
	word used;

#ifdef MOXA_DEBUG_TIMER
	CP('T');
#endif

	del_timer(&mx_timer);

	for(chan=0;chan<num_channels;chan++) {
		info = &mx_port[chan];
		ctrl = info->ctrl;
		if (info->write_in_progress) {
			save_flags(flags); cli();
			used = (ctrl->tx_queue_write - ctrl->tx_queue_read);
			if (!used)
				info->write_in_progress = 0;
			restore_flags(flags);
			used &= (MOXA_TX_BUF_SIZE - 1);
			if (used < MOXA_TX_THRESH)
				mx_sched_event(info, MOXA_SWEVENT_TBE);
		}
		if (ctrl->rx_queue_write != ctrl->rx_queue_read)
			mx_sched_event(info, MOXA_SWEVENT_READ);
	}
	
	mx_timer.expires = jiffies + MOXA_TIMER_PERIOD;
	add_timer(&mx_timer);
}

/*
 *	Card interrupt service routine. Called whenever an event occurs
 *	on any card in the system.
 */

static void
mx_interrupt(int irq, void *devid, struct pt_regs *regs)
{
	struct moxa_event_ctrl *events;
	struct moxa_port *info;
	word *chan_events;
	word evts;
	int card;
	int chan;
	int retry;

#ifdef MOXA_DEBUG_EVENTS
	CP('I');
#endif

	/*
	 * We will check all the channels to see which of them require servicing.
	 * Only event detection is performed -- the event processing itself
	 * is done by the bottom half of the driver. The whole cycle stops
	 * when there is nothing more to do.
	 */

	do {
		retry = 0;
		for (card=0; card<4; card++)
			if ((events = mx_events[card])) {
				if (events->int_flag) {
					chan_events = (word *) (moxa_addr[card] + 8 +
								events -> event_table_offset);
					for(chan=0; chan<8; chan++) {
						if ((evts = *chan_events)) {
#ifdef MOXA_DEBUG_EVENTS
							CP('e'); CP1(card); CP1(chan); CP2(evts);
#endif
							info = &mx_port[8*card + chan];
							if (evts & MOXA_EVENT_COUNT)
								mx_sched_event(info, MOXA_SWEVENT_READ);
							if (evts & MOXA_EVENT_BREAK)
								mx_sched_event(info, MOXA_SWEVENT_BREAK);
							if (evts & MOXA_EVENT_MODEM)
								mx_sched_event(info, MOXA_SWEVENT_MODEM);
							*chan_events=0;
						}
						chan_events++;
					}
					retry = 1;
					events->int_flag = 0;
				}
			}
	}
	while (retry);
#ifdef MOXA_DEBUG_EVENTS
	CP(']');
#endif
}

/*
 * This routine is used to handle the "bottom half" processing for the
 * serial driver, known also the "software interrupt" processing.
 * This processing is done at the kernel interrupt level, after the
 * mx_interrupt() has returned, BUT WITH INTERRUPTS TURNED ON.  This
 * is where time-consuming activities which can not be done in the
 * interrupt driver proper are done; the interrupt driver schedules
 * them using mx_sched_event(), and they get done here.
 *
 * This is done through one level of indirection--the task queue.
 * When a hardware interrupt service routine wants service by the
 * driver's bottom half, it enqueues the appropriate tq_struct (one
 * per port) to the system task queue "tq_immediate" and requests
 * processing of this queue later (i.e., after the interrupt returns).
 */

static void
do_softint(void *private)
{
	struct moxa_port *info = (struct moxa_port *) private;
	volatile struct moxa_chan_ctrl *ctrl = info->ctrl;
	struct tty_struct *tty = info->tty;
	ulg flags;
	int events;
	byte ch;

#ifdef MOXA_DEBUG_EVENTS
	CP('S'); CP2(info->line);
#endif

	/*
	 * Get and reset event mask.
	 */

	save_flags(flags); cli();
	events = info->event;
	info->event = 0;
	restore_flags(flags);

	/*
	 * First of all, store incoming characters into the flip buffer.
	 */

	if (events & MOXA_SWEVENT_READ) {
		if (!tty) {			/* No tty -> discard all characters */
#ifdef MOXA_DEBUG_EVENTS
			CP('r');
#endif
			ctrl->rx_queue_read = ctrl->rx_queue_write;
		} else {
#ifdef MOXA_DEBUG_EVENTS
			CP('R');
#endif
			while (tty->flip.count < TTY_FLIPBUF_SIZE &&
			       ctrl->rx_queue_read != ctrl->rx_queue_write) {
				save_flags(flags); cli();
				ch = mx_get_byte(info, ctrl->rx_queue_read + info->read_buffer);
				restore_flags(flags);
				tty->flip.count++;
				*tty->flip.flag_buf_ptr++ = TTY_NORMAL;
				*tty->flip.char_buf_ptr++ = ch;
				ctrl->rx_queue_read =
					(ctrl->rx_queue_read + 1) & (MOXA_RX_BUF_SIZE - 1);
			}
#ifdef MOXA_DEBUG_OVERFLOW
			if (tty->flip.count == TTY_FLIPBUF_SIZE)
				printk("moxa: flip buffer overflow\n");
#endif
			queue_task(&tty->flip.tqueue, &tq_timer);
		}

#ifdef MOXA_DEBUG_OVERFLOW
		if (ctrl->status & MOXA_STAT_OVERFLOW) {
			printk("moxa: BUFFER OVERFLOW !!\n");
		}
		if (ctrl->status & MOXA_STAT_FULL) {
			printk("moxa: BUFFER FULL !!\n");
		}
#endif
			
		if ((((ctrl->rx_queue_write - ctrl->rx_queue_read)
		      & (MOXA_RX_BUF_SIZE - 1)) < (MOXA_RX_BUF_SIZE - MOXA_RX_WAKEUP_LEN))
		    && (ctrl->status & MOXA_STAT_FULL)) {   /* We need to wake up receiver handshaking */
#ifdef MOXA_DEBUG_EVENTS
			CP('w');
#endif
			info->receiver_wakeup = 1;
			events |= MOXA_SWEVENT_MODEM;
		}
	}

	/*
	 * Then wake up sleeping write requests when there's enough room.
	 */

	if (events & MOXA_SWEVENT_TBE) {
#ifdef MOXA_DEBUG_EVENTS
		CP('W');
#endif
		info->ctrl->start_xmit = 1;
		if (tty) {
			if ((tty->flags & (1<<TTY_DO_WRITE_WAKEUP)) &&
			    tty->ldisc.write_wakeup)
				(tty->ldisc.write_wakeup)(tty);
			wake_up_interruptible(&tty->write_wait);
		}
	}

	/*
	 * Check UART BREAK event.
	 */

	if (events & MOXA_SWEVENT_BREAK) {
#ifdef MOXA_DEBUG_EVENTS
		CP('B');
#endif
		if (tty && tty->flip.count < TTY_FLIPBUF_SIZE) {
			tty->flip.count ++;
			*tty->flip.flag_buf_ptr++ = TTY_BREAK;
			*tty->flip.char_buf_ptr++ = 0;
		}
	}

	/*
	 * If we wrote anything to the flip buffer, wake up the ldisc.
	 */

	if (tty && tty->flip.count)
		queue_task(&tty->flip.tqueue, &tq_timer);

	/*
	 * Check modem status changes. This is done in a special way, because
	 * it requires sending of MOXA command, which may already be in progress.
	 */

	if (events & MOXA_SWEVENT_MODEM) {
#ifdef MOXA_DEBUG_EVENTS
		CP('m');
#endif
		schedule_modem_event(info);
	}
}

/*
 *	Process `modem status changed' event. Called by get_modem_status
 *	when the event has been scheduled by interrupt driver and the
 *	status of the modem lines has been read successfully.
 *
 *	This process is also used in somewhat tricky way to kick receiver
 *	handshaking if enough room in the buffer is free.
 */

static void
modem_event(struct moxa_port *info)
{
#ifdef MOXA_DEBUG_EVENTS
	CP('M');
#endif

	if (info->receiver_wakeup)
		mx_cmd(info, MOXA_CMD_RESUME);

	if (info->flags & ASYNC_CHECK_CD) {
		if (!(info->modem_status_old & MOXA_MSTAT_DCD)
		    && (info->modem_status & MOXA_MSTAT_DCD)) {
			DBPRINT(("OPEN_WAKEUP "));
		  	wake_up_interruptible(&info->open_wait);
		} else if ((info->modem_status_old & MOXA_MSTAT_DCD)
			   && !(info->modem_status & MOXA_MSTAT_DCD)
			   && !((info->flags & ASYNC_CALLOUT_ACTIVE)
				&& (info->flags & ASYNC_CALLOUT_NOHUP))) {
			DBPRINT(("HANGUP "));
			tty_hangup(info->tty);
			wake_up_interruptible(&info->open_wait);
			info->flags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CALLOUT_ACTIVE);
		}
	}
}

/*
 *	This is called whenever a port becomes active;
 *	interrupts are enabled and DTR & RTS are turned on.
 */

static int
startup(struct moxa_port * info)
{
	if (info->flags & ASYNC_INITIALIZED)
		return 0;

	if (!info->type) {
		DBPRINT(("No info->type "));
		if (info->tty)
			set_bit(TTY_IO_ERROR, &info->tty->flags);
		return 0;
	}
	
	mx_cmd1(info,MOXA_CMD_RESET_CHAN,MOXA_MAGIC_COOKIE);
	mx_cmd(info,MOXA_CMD_INIT_CHAN);
	mx_cmd1(info,MOXA_CMD_OPEN,MOXA_MAGIC_COOKIE);
	mx_cmd1(info,MOXA_CMD_COUNT_CTRL,1);
		
	info->modem_ctrl = MOXA_MCTRL_DTR | MOXA_MCTRL_RTS;
	info->modem_ctrl_old = 0xffff;	/* Force new setting of modem lines */
	config_setup(info);

	info->flags |= ASYNC_INITIALIZED;
	if (info->tty)
		clear_bit(TTY_IO_ERROR, &info->tty->flags);

	return 0;
} 

static inline void
start_xmit(struct moxa_port *info )
{
	DBPRINT(("start_xmit "));
	info->write_in_progress = 1;
	info->ctrl->start_xmit = 1;
} 

/*
 *	Send changed modem control signals to the card.
 */

static inline void
update_modem_ctrl(struct moxa_port * info)
{
	if (info->modem_ctrl != info->modem_ctrl_old) {
		mx_cmd1(info, MOXA_CMD_SET_MODEM, info->modem_ctrl);
		info->modem_ctrl_old = info->modem_ctrl;
	}
}

/*
 *	Set / reset modem event requests.
 */

static void
set_modem_events(struct moxa_port *info, int control)
{
	if (info->modem_events_enabled && !control)
		mx_cmd(info, MOXA_CMD_DISABLE_MODEM_EVENTS);
	else if (!info->modem_events_enabled && control) {
		mx_cmd(info, MOXA_CMD_MODEM_EVENT_ON);
		schedule_modem_event(info);
	}
	info->modem_events_enabled = control;
}

/*
 *	This routine shuts down a serial port; event interrupts are disabled,
 *	and DTR is dropped if the hangup on close termio flag is on.
 */

static void
shutdown(struct moxa_port * info)
{
	DBPRINT(("shutting down "));

	if (!(info->flags & ASYNC_INITIALIZED))
		return;

	if (!info->tty || (info->tty->termios->c_cflag & HUPCL)) {
		DBPRINT(("hanging up "));
		info->modem_ctrl &= ~MOXA_MCTRL_DTR;
	}

	if (info->tty)
		set_bit(TTY_IO_ERROR, &info->tty->flags);
	info->flags &= ~ASYNC_INITIALIZED;

	update_modem_ctrl(info);
	mx_cmd1(info,MOXA_CMD_COUNT_CTRL,0);
	set_modem_events(info, 0);
/* XXX */ /* mx_cmd1(info,MOXA_CMD_CLOSE,MOXA_MAGIC_COOKIE); */

	DBPRINT(("shutdown done "));
	return;
}

/*
 *	This routine finds or computes the various line characteristics.
 */

static void
config_setup(struct moxa_port * info)
{
	ulg cflag;
	int i;

	if (!info->tty || !info->tty->termios)
		return;
	if (info->line == -1)
		return;

	cflag = info->tty->termios->c_cflag;

	/*
	 * Set baud rate, word length, stop bits and parity.
	 */

	i = cflag & CBAUD;
	if (i & CBAUDEX) {
		if (i == B57600)
			i = 16;
		else if (i == B115200) 
			i = 17;
		else
			info->tty->termios->c_cflag &= ~CBAUDEX;
	}
	if (i == 15) {
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
			i += 1;
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
			i += 3;
	}

	if (i) {
		if (baud_table[i] >= 0)
			mx_cmd1(info,MOXA_CMD_SET_BAUD,baud_table[i]);
		else
			printk(KERN_WARNING "moxa: unsupported baud rate #%d\n",i);
	}

	/*
         * By tradition (is it a standard?) a baud rate of zero
	 * implies the line should be/has been closed.  A bit
	 * later in this routine such a test is performed.
	 */

	switch (cflag & CSIZE) {
	case CS5:
		i = MOXA_MODE_LEN_5;
		break;
	case CS6:
		i = MOXA_MODE_LEN_6;
		break;
	case CS7:
		i = MOXA_MODE_LEN_7;
		break;
	case CS8:
		i = MOXA_MODE_LEN_8;
		break;
	}

	if (cflag & CSTOPB)
		i |= MOXA_MODE_STOP_2;

	if (cflag & PARENB)
		if (cflag & PARODD)
			i |= MOXA_MODE_PAR_ODD;
		else
			i |= MOXA_MODE_PAR_EVEN;

	mx_cmd1(info,MOXA_CMD_SET_MODE,i);

	/*
	 * Set flow control flags.
	 */

	if (cflag & CRTSCTS)
		mx_cmd1(info, MOXA_CMD_SET_FLOW, MOXA_FLOW_CTS | MOXA_FLOW_RTS);
	else
		mx_cmd1(info, MOXA_CMD_SET_FLOW, 0);

	if (cflag & CLOCAL) {
		info->flags &= ~ASYNC_CHECK_CD;
		set_modem_events(info, 0);
	} else {
		info->flags |= ASYNC_CHECK_CD;
		set_modem_events(info, 1);
	}

	/*
	 * If the baud rate is zero, turn off the line (drop DTR).
	 */

	if ((cflag & CBAUD) == 0) {
		DBPRINT(("hanging up "));
		info->modem_ctrl &= ~MOXA_MCTRL_DTR;
	} else
		info->modem_ctrl |= MOXA_MCTRL_DTR;

	update_modem_ctrl(info);

	if (info->tty)
		clear_bit(TTY_IO_ERROR, &info->tty->flags);
}

/*
 *	mx_start and mx_stop provice software output flow control
 *	(XON/XOFF, software CTS etc.). Not implemented in this version,
 *	because the handshaking is processed automatically by the hardware.
 */

static void
mx_stop(struct tty_struct *tty)
{
	DBPRINT(("mx_stop "));
}

static void
mx_start(struct tty_struct *tty)
{
	DBPRINT(("mx_start "));
}

/*
 *	mx_put_char is called to store one character to the write queue.
 *	The writing itself is not started -- mx_flush_chars has to be called
 *	later to kick the card.
 */

static void
mx_put_char(struct tty_struct *tty, byte ch)
{
	struct moxa_port *info = (struct moxa_port *)tty->driver_data;
	volatile struct moxa_chan_ctrl *ctrl;
	ulg flags;
	word newwrite;

	DBPRINT(("mx_put_char "));
	if (serial_paranoia_check(info, tty->device, "mx_put_char"))
		return;

	ctrl = info->ctrl;
	save_flags(flags); cli();
	newwrite = (ctrl->tx_queue_write + 1) & (MOXA_TX_BUF_SIZE - 1);
	if (newwrite != ctrl->tx_queue_read) {
		mx_put_byte(info, info->write_buffer + ctrl->tx_queue_write, ch);
		ctrl->tx_queue_write = newwrite;
	} else {
#ifdef MOXA_DEBUG_OVERFLOW
		printk("moxa: mx_put_char buffer overflow !\n");
#endif
	}
	restore_flags(flags);
}

/*
 *	mx_flush_chars starts real transmit of all characters queued by
 *	mx_put_char.
 */

static void
mx_flush_chars(struct tty_struct *tty)
{
	struct moxa_port *info = (struct moxa_port *)tty->driver_data;

	DBPRINT(("mx_flush_chars "));
	if (serial_paranoia_check(info, tty->device, "mx_flush_chars"))
		return;

	if (tty->stopped || tty->hw_stopped)
		return;

	start_xmit(info);
}

/*
 *	This routine gets called when the tty driver wants to send
 *	more than a single character.
 */

static int
mx_write(struct tty_struct * tty, int from_user, const byte *buf, int count)
{
	struct moxa_port *info = (struct moxa_port *)tty->driver_data;
	volatile struct moxa_chan_ctrl *ctrl;
	ulg flags;
	int c,done;
	int total = 0;
	word newwrite;
	const byte *ch;

	DBPRINT(("mx_write(%d) ", count));
	if (serial_paranoia_check(info, tty->device, "mx_write"))
		return 0;

	ctrl = info->ctrl;

	for (;;) {
		save_flags(flags); cli();
		c = (info->ctrl->tx_queue_write - info->ctrl->tx_queue_read);
		restore_flags(flags);
		c = (MOXA_TX_BUF_SIZE-1) - (c & (MOXA_TX_BUF_SIZE-1));
		c = MIN(c, count);
		if (!c)
			break;

		if (from_user) {
			down(&tmp_buf_sem);
			c = MIN(c, PAGE_SIZE);
			if (copy_from_user(tmp_buf, buf, c))
				return -EINVAL;
			ch = tmp_buf;
		}
		else
			ch=buf;
		done = 0;

		while (done < c) {
			save_flags(flags); cli();
			newwrite = (ctrl->tx_queue_write + 1) & (MOXA_TX_BUF_SIZE - 1);
			if (newwrite != ctrl->tx_queue_read) {
				mx_put_byte(info, info->write_buffer + ctrl->tx_queue_write, *ch);
				ctrl->tx_queue_write = newwrite;
				restore_flags(flags);
			} else {
				restore_flags(flags);
				break;
			}
			done++;
			ch++;
		}

		if (from_user)
			up(&tmp_buf_sem);

		total += done;
		count -= done;
		buf += done;
	}

	if (!tty->stopped && !tty->hw_stopped)
		start_xmit(info);

	return total;
}

/*
 *	mx_write_room returns how much free space is in the write buffer.
 */

static int
mx_write_room(struct tty_struct *tty)
{
	struct moxa_port *info = (struct moxa_port *)tty->driver_data;
	ulg flags;
	int space;

	DBPRINT(("mx_write_room"));
	if (serial_paranoia_check(info, tty->device, "mx_write_room"))
		return 0;

	save_flags(flags); cli();
	space = (info->ctrl->tx_queue_write - info->ctrl->tx_queue_read);
	restore_flags(flags);
	space = (MOXA_TX_BUF_SIZE-1) - (space & (MOXA_TX_BUF_SIZE-1));

	DBPRINT(("=%d ", space));
	return space;
}

/*
 *	mx_chars_in_buffer returns how many characters are ready in RX buffer.
 */

static int
mx_chars_in_buffer(struct tty_struct *tty)
{
	struct moxa_port *info = (struct moxa_port *)tty->driver_data;
	ulg flags;
	int space;

	DBPRINT(("mx_chars_in_buffer"));
	if (serial_paranoia_check(info, tty->device, "mx_chars_in_buffer"))
		return 0;

	save_flags(flags); cli();
	space = (info->ctrl->rx_queue_write - info->ctrl->rx_queue_read)
		& (MOXA_RX_BUF_SIZE - 1);
	restore_flags(flags);

	DBPRINT(("=%d ", space));
	return space;
}

/*
 *	mx_flush_buffer trashes the whole transmit buffer.
 */

static void
mx_flush_buffer(struct tty_struct *tty)
{
	struct moxa_port *info = (struct moxa_port *)tty->driver_data;
	
	DBPRINT(("mx_flush_buffer "));
	if (serial_paranoia_check(info, tty->device, "mx_flush_buffer"))
		return;

#if 0
	mx_cmd1(info, MOXA_CMD_CLEAR, MOXA_CLEAR_TX);	    /* XXX: FIXME */

	wake_up_interruptible(&tty->write_wait);
	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup)(tty);
#endif
}

/*
 *	IOCTL's
 */

static int
get_modem_info(struct moxa_port * info, unsigned int *value)
{
	word status;
	unsigned int result;

	DBPRINT(("get_modem_info "));

	get_modem_status(info);
	status = info->modem_status;

	result = ((status & MOXA_MSTAT_CTS) ? TIOCM_CTS : 0)
		| ((status & MOXA_MSTAT_DSR) ? TIOCM_DSR : 0)
		| ((status & MOXA_MSTAT_RING) ? TIOCM_RNG : 0)
		| ((status & MOXA_MSTAT_DCD) ? TIOCM_CAR : 0)
		| ((info->modem_ctrl & MOXA_MCTRL_DTR) ? TIOCM_DTR : 0)
		| ((info->modem_ctrl & MOXA_MCTRL_RTS) ? TIOCM_RTS : 0);
	if (put_user(result, value))
		return -EINVAL;

	DBPRINT(("->%x",result));	
	return 0;
}

static int
set_modem_info(struct moxa_port * info, unsigned int cmd,
	       unsigned int *value)
{
	unsigned int arg;

	if (get_user(arg, value))
		return -EINVAL;

	switch (cmd) {
	case TIOCMBIS:
		DBPRINT(("TIOCMBIS(%x) ",arg));
		if (arg & TIOCM_RTS)
			info->modem_ctrl |= MOXA_MCTRL_RTS;
		if (arg & TIOCM_DTR)
			info->modem_ctrl |= MOXA_MCTRL_DTR;
		break;
	case TIOCMBIC:
		DBPRINT(("TIOCMBIC(%x) ",arg));
		if (arg & TIOCM_RTS)
			info->modem_ctrl &= ~MOXA_MCTRL_RTS;
		if (arg & TIOCM_DTR)
			info->modem_ctrl &= ~MOXA_MCTRL_DTR;
		break;
	case TIOCMSET:
		DBPRINT(("TIOCMSET(%x) ",arg));
		if (arg & TIOCM_RTS)
			info->modem_ctrl |= MOXA_MCTRL_RTS;
		else
			info->modem_ctrl &= ~MOXA_MCTRL_RTS;
		if (arg & TIOCM_DTR)
			info->modem_ctrl |= MOXA_MCTRL_DTR;
		else
			info->modem_ctrl &= ~MOXA_MCTRL_DTR;
		break;
	default:
		return -EINVAL;
	}
	update_modem_ctrl(info);
	return 0;
}

static int
get_serial_info(struct moxa_port *info, struct serial_struct *retinfo)
{
	struct serial_struct tmp;

	DBPRINT(("get_serial_info "));
	if (!retinfo)
		return -EFAULT;
	memset(&tmp, 0, sizeof(tmp));
	tmp.type = info->type;
	tmp.line = info->line;
	tmp.port = (int) info->ctrl;
	tmp.irq = MOXA_IRQ;
	tmp.flags = info->flags;
	tmp.close_delay = info->close_delay;
	if (copy_to_user(retinfo, &tmp, sizeof(tmp)))
		return -EINVAL;
	DBPRINT(("okay "));
	return 0;
}

static int
set_serial_info(struct moxa_port *info, struct serial_struct *new_info)
{
	struct serial_struct new;

	DBPRINT(("set_serial_info "));
	if (!new_info)
		return -EFAULT;
	if (copy_from_user(&new, new_info, sizeof(new)))
		return -EINVAL;

	if (!suser()) {
		if ((new.close_delay != info->close_delay)
		    || ((new.flags ^ info->flags) & ASYNC_FLAGS & ~ASYNC_USR_MASK))
		 	return -EPERM;
		info->flags = ((info->flags & ~ASYNC_USR_MASK)
			       | (new.flags & ASYNC_USR_MASK));
	} else {
		info->flags = (info->flags & ~ASYNC_FLAGS) | (new.flags & ASYNC_FLAGS);
		info->close_delay = new.close_delay;
	}

	if (info->flags & ASYNC_INITIALIZED) {
		config_setup(info);
		return 0;
	} else
		return startup(info);
}

#ifdef MOXA_DEBUGGING_IOCTL

static void
moxa_debug(void)
{
	int i;
	struct moxa_port *p;

	for(i=0; i < NR_PORTS; i++) {
		p = mx_port + i;
		if (p->line < 0)
			continue;
		printk("Port %d: flags=%x pgport=%x memwin=%x rbuf=%lx wbuf=%lx line=%d count=%d block=%d \
closedly=%d sess=%ld pgrp=%ld\n",
		       i, p->flags, (int)p->page_port, (int)p->mem_window, p->read_buffer, p->write_buffer,
		       p->line, p->count, p->blocked_open, p->close_delay, p->session, p->pgrp);
		printk("\tevent=%x mctrl=%x mctrl0=%x mstat=%x mstat0=%x mes=%d cip=%d \
wip=%d rwa=%d men=%d rqr=%x rqw=%x tqr=%x tqw=%x\n",
		       p->event, p->modem_ctrl, p->modem_ctrl_old, p->modem_status, p->modem_status_old,
		       p->modem_event_scheduled, p->cmd_in_progress, p->write_in_progress, p->receiver_wakeup,
		       p->modem_events_enabled, p->ctrl->rx_queue_read, p->ctrl->rx_queue_write,
		       p->ctrl->tx_queue_read, p->ctrl->tx_queue_write);
		printk("\tf1=%x stat=%x flowc=%x brkc=%x sxmit=%x cmd=%x par1=%x par2=%x par3=%x\n",
		       p->ctrl->flags1, p->ctrl->status,
		       p->ctrl->flowctrl, p->ctrl->break_cnt, p->ctrl->start_xmit, p->ctrl->command,
		       p->ctrl->par1, p->ctrl->par2, p->ctrl->par3);
	}
}

#endif

static int
mx_ioctl(struct tty_struct *tty, struct file * file,
	 unsigned int cmd, unsigned long arg)
{
	int x;
	struct moxa_port * info = (struct moxa_port *)tty->driver_data;
	int ret_val = 0;

	DBPRINT(("mx_ioctl ttyM%d, cmd = %x arg = %lx: ", info->line, cmd, arg));
	if (serial_paranoia_check(info, tty->device, "mx_ioctl"))
		return -EFAULT;

	switch (cmd) {
	case TIOCMBIS:
	case TIOCMBIC:
	case TIOCMSET:
		ret_val = set_modem_info(info, cmd, (unsigned int *) arg);
		break;

/* The following commands are incompletely implemented!!! */
	case TIOCGSOFTCAR:
		if (put_user(!!C_CLOCAL(tty), (int *) arg))
			return -EINVAL;
		break;
			
	case TIOCSSOFTCAR:
		if (get_user(x, (int *) arg))
			return -EINVAL;
		tty->termios->c_cflag =
			((tty->termios->c_cflag & ~CLOCAL) |
			 (x ? CLOCAL : 0));
		break;
			
	case TIOCMGET:
		ret_val = get_modem_info(info, (unsigned int *) arg);
		break;

	case TIOCGSERIAL:
		ret_val = get_serial_info(info,(struct serial_struct *) arg);
		break;
			
	case TIOCSSERIAL:
		ret_val = set_serial_info(info,(struct serial_struct *) arg);
		break;

#ifdef MOXA_DEBUGGING_IOCTL
	case 0xdead:
		moxa_debug();
		ret_val = 0;
		break;
#endif
			
	default:
		DBPRINT(("unknown"));
		ret_val = -ENOIOCTLCMD;
	}

	DBPRINT(("\n"));
	return ret_val;
}

/*
 *	Called when new termios are set to notify the driver what's
 *	being done.
 */

static void
mx_set_termios(struct tty_struct *tty, struct termios * old_termios)
{
	struct moxa_port *info = (struct moxa_port *)tty->driver_data;

	DBPRINT(("mx_set_termios "));
	if (serial_paranoia_check(info, tty->device, "mx_set_termios"))
		return;
	
	if (tty->termios->c_cflag == old_termios->c_cflag)
		return;

	config_setup(info);

	if (!(old_termios->c_cflag & CLOCAL) &&
	    (tty->termios->c_cflag & CLOCAL))
		wake_up_interruptible(&info->open_wait);

	return;
}

/*
 *	Close the serial port.
 */

static void
mx_close(struct tty_struct * tty, struct file * filp)
{
	struct moxa_port * info = (struct moxa_port *)tty->driver_data;

	DBPRINT(("mx_close %d: ",info->line));
	if (!info || serial_paranoia_check(info, tty->device, "mx_close"))
		return;

	if ((tty->count == 1) && (info->count != 1)) {
	/*
	 * Uh, oh.  tty->count is 1, which means that the tty
	 * structure will be freed.  Info->count should always
	 * be one in these conditions.  If it's greater than
	 * one, we've got real problems, since it means the
	 * serial port won't be shutdown.
	 */
		printk("mx_close: bad serial port count; tty->count is 1, "
		       "info->count is %d\n", info->count);
		info->count = 1;
	}
		
	if (--info->count < 0) {
		printk("mx_close: bad serial port count for ttys%d: %d\n",
		       info->line, info->count);
		info->count = 0;
	}
		
	MOD_DEC_USE_COUNT;
	if (info->count) {
		DBPRINT(("Not yet\n"));
		return;
	}
	info->flags |= ASYNC_CLOSING;
	
	/*
	 * Save the termios structure, since this port may have
	 * separate termios for callout and dialin.
	 */
	 
	if (info->flags & ASYNC_NORMAL_ACTIVE)
		info->normal_termios = *tty->termios;
	if (info->flags & ASYNC_CALLOUT_ACTIVE)
		info->callout_termios = *tty->termios;
	if (info->flags & ASYNC_INITIALIZED) {
		tty_wait_until_sent(tty, 3000); /* 30 seconds timeout */
		shutdown(info);
	}
	if (tty->driver.flush_buffer)
		tty->driver.flush_buffer(tty);
	if (tty->ldisc.flush_buffer)
		tty->ldisc.flush_buffer(tty);
	info->event = 0;
	info->tty = 0;

	if (info->blocked_open) {
		if (info->close_delay) {
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(jiffies);
		}
		wake_up_interruptible(&info->open_wait);
	}
	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE|ASYNC_CLOSING);
	wake_up_interruptible(&info->close_wait);

	DBPRINT(("closed.\n"));
	return;
}

/*
 *	mx_hangup() --- called by tty_hangup() when a hangup is signaled.
 */

static void
mx_hangup(struct tty_struct *tty)
{
	struct moxa_port * info = (struct moxa_port *)tty->driver_data;

	DBPRINT(("mx_hangup "));
	if (serial_paranoia_check(info, tty->device, "mx_hangup"))
		return;
    
	shutdown(info);
	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE);
	wake_up_interruptible(&info->open_wait);
}

/*
 *	Block the called until selected port is free and ready for use.
 */

static int
block_til_ready(struct tty_struct *tty, struct file * filp,
		struct moxa_port *info)
{
	struct wait_queue wait = { current, NULL };
	ulg flags;
	int retval;
	word state;

	DBPRINT(("block_til_ready "));

	/*
	 * If the device is in the middle of being closed, then block
	 * until it's done, and then try again.
	 */
	 
	if (info->flags & ASYNC_CLOSING) {
		interruptible_sleep_on(&info->close_wait);
		if (info->flags & ASYNC_HUP_NOTIFY)
			return -EAGAIN;
		else
			return -ERESTARTSYS;
	}

	/*
	 * If this is a callout device, then just make sure the normal
	 * device isn't being used.
	 */

	if (tty->driver.subtype == SERIAL_TYPE_CALLOUT) {
		if (info->flags & ASYNC_NORMAL_ACTIVE)
			return -EBUSY;
		
		if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    (info->flags & ASYNC_SESSION_LOCKOUT) &&
		    (info->session != current->session))
			return -EBUSY;
		 
		if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    (info->flags & ASYNC_PGRP_LOCKOUT) &&
		    (info->pgrp != current->pgrp))
			return -EBUSY;

		info->flags |= ASYNC_CALLOUT_ACTIVE;
		return 0;
	}

	/*
	 * If non-blocking mode is set, then make the check up front
	 * and then exit.
	 */

	if (filp->f_flags & O_NONBLOCK) {
		if (info->flags & ASYNC_CALLOUT_ACTIVE)
			return -EBUSY;
		info->flags |= ASYNC_NORMAL_ACTIVE;
		return 0;
	}

	/*
	 * Block waiting for the carrier detect and the line to become
	 * free (i.e., not in use by the callout).  While we are in
	 * this loop, info->count is dropped by one, so that
	 * mx_close() knows when to free things.  We restore it upon
	 * exit, either normal or abnormal.
	 */

	retval = 0;
	add_wait_queue(&info->open_wait, &wait);
	DBPRINT(("waiting for /DCD "));
	info->count--;
	info->blocked_open++;

	while (1) {
		save_flags(flags); cli();
		if (!(info->flags & ASYNC_CALLOUT_ACTIVE)) {
			info->modem_ctrl |= MOXA_MCTRL_DTR;
		}
		restore_flags(flags);
		current->state = TASK_INTERRUPTIBLE;
		if (tty_hung_up_p(filp)	|| !(info->flags & ASYNC_INITIALIZED) ) {
			if (info->flags & ASYNC_HUP_NOTIFY)
				retval = -EAGAIN;
			else
				retval = -ERESTARTSYS;
			break;
		}
		get_modem_status(info);
		state = info->modem_status;
		save_flags(flags); cli();
		if (!(info->flags & ASYNC_CALLOUT_ACTIVE)
		    && !(info->flags & ASYNC_CLOSING)
		    && (C_CLOCAL(tty)
			|| (state & MOXA_MSTAT_DCD))) {
			restore_flags(flags);
			break;
		}
		restore_flags(flags);
		if (signal_pending(current)) {
			retval = -ERESTARTSYS;
			break;
		}
		DBPRINT(("blocking "));
		schedule();
	}
		
	current->state = TASK_RUNNING;
	remove_wait_queue(&info->open_wait, &wait);
	if (!tty_hung_up_p(filp))
		info->count++;
	info->blocked_open--;
	DBPRINT(("afterblock "));
	if (retval)
		return retval;
	info->flags |= ASYNC_NORMAL_ACTIVE;
	return 0;
}

/*
 *	This routine is called whenever a serial port is opened.  It
 *	performs the serial-specific initialization for the tty structure.
 */

int
mx_open(struct tty_struct *tty, struct file *filp)
{
	struct moxa_port  *info;
	int retval, line;

	DBPRINT(("mx_open..."));

	line = MINOR(tty->device) - tty->driver.minor_start;
	if ((line < 0) || (NR_PORTS <= line)) {
		DBPRINT(("invalid minor\n"));
		return -ENODEV;
	}
	info = &mx_port[line];
	if (info->line < 0 || info->magic == -1) {
		DBPRINT(("channel doesn't exist\n"));
		return -ENODEV;
	}
	if (serial_paranoia_check(info, tty->device, "mx_open"))
		return -ENODEV;
	DBPRINT(("channel %d ",info->line));
	
	MOD_INC_USE_COUNT;
	info->count++;
	tty->driver_data = info;
	info->tty = tty;

	if (!tmp_buf) {
		tmp_buf = (byte *) get_free_page(GFP_KERNEL);
		if (!tmp_buf) {
			DBPRINT(("No memory for tmpbuf\n"));
			MOD_DEC_USE_COUNT;
			info->count--;
			return -ENOMEM;
		}
	}

	if ((info->count == 1) && (info->flags & ASYNC_SPLIT_TERMIOS)) {
		if (tty->driver.subtype == SERIAL_TYPE_NORMAL)
			*tty->termios = info->normal_termios;
		else 
			*tty->termios = info->callout_termios;
	}
		
	/*
	 * Start up serial port
	 */

	DBPRINT(("startup "));
	retval = startup(info);
	if (retval) {
		DBPRINT(("startup failed\n"));
		MOD_DEC_USE_COUNT;
		info->count--;
		return retval;
	}

	retval = block_til_ready(tty, filp, info);
	if (retval) {
		DBPRINT(("b_t_r failed\n"));
		MOD_DEC_USE_COUNT;
		info->count--;
		return retval;
	}

	info->session = current->session;
	info->pgrp = current->pgrp;

	DBPRINT(("OK\n"));
	return 0;
}

/*
 *	Card ROM contents (linked to the module)
 */

extern int rom_len;
extern byte rom_start[];

/*
 *	Wait for card handshaking during init.
 */

static int
mx_wait_handshake(volatile word *shake,int value)
{
	int timeout = jiffies + 10*HZ;
	int j,k;

	while (jiffies < timeout) {
		if (*shake == value) return 1;
		small_delay(50);	/* We shouldn't hog the shared memory bus */
	}
	printk("Handshaking timeout.\n");
	return 0;
}

#define WAIT_HANDSHAKE if (!mx_wait_handshake(handshake,0x218)) return 0;
#define WAIT_HANDSHAKE2 if (!mx_wait_handshake(handshake2,0x404)) return 0;
#define SEND_HANDSHAKE *handshake = 0;

/*
 *	Card hardware initialization.
 */

static int
mx_card_init(byte *address)
{
	volatile byte *page_port = address + 0x1ff0;
	word *area_0000 = (word *) address;
	word *area_2000 = (word *) (address + 0x2000);
	word *area_3000 = (word *) (address + 0x3000);
	volatile unsigned int *test_point = (unsigned int *) area_2000;
	volatile word *handshake = (word *) (address + 0x2004);
	volatile word *handshake2 = (word *) (address + 0x0108);
	volatile word *error = (word *) (address + 0x2002);
	int err;
	int i;
	byte sum;

	/*
	 * First of all, assert MOXA CPU reset, clear and test presence
	 * of card memory at given address.
	 */

	*page_port = 0xbf;				    /* Assert RESET and use last mem. page */

	*test_point = 0xdeadf00d;
	test_point[0x100] = 0;
	if (*test_point != 0xdeadf00d) {
		printk("Card memory not found!\n");
		return 0;
	}
	*test_point = 0x12349876;
	test_point[0x100] = 0;
	if (*test_point != 0x12349876) {
		printk("Card memory test failed!\n");
		return 0;
	}

	memset(area_2000,0,0x2000);

	/*
	 * Copy hardware test to card RAM and run it.
	 */
	
	memcpy(area_3000, rom_start, 0x1000);
	*page_port = 0x3f;				    /* Release RESET */

	printk(".");
	WAIT_HANDSHAKE;

	if ((err = *error)) {
		if (err & 0x0100) printk(" RAM1 error");
		if (err & 0x0200) printk(" RAM2 error");
		if (err & 0x0400) printk(" Hardware error");
		if (err & 0x0800) printk(" Unexpected interrupt");
		if (err & 0xf000) printk(" Unidentified error");
		return 0;
	}

	/*
	 * Calculate checksum of the main portion of the driver and install it.
	 */

	sum = 0;
	for (i=0x1000; i<rom_len; i++)
		sum += rom_start[i];
	
	memcpy(area_0000, rom_start+0x1000, rom_len-0x1000);
	area_2000[3] = (rom_len-0x1000) >> 1;

	SEND_HANDSHAKE;
	printk(".");
	WAIT_HANDSHAKE;

	area_2000[4] = sum;
	SEND_HANDSHAKE;
	printk(".");
	WAIT_HANDSHAKE;
	if (!area_2000[5]) {
		printk(" Transfer error");
		return 0;
	}

	/*
	 * Start the driver and wait until it's ready.
	 */

	SEND_HANDSHAKE;
	printk(".");
	WAIT_HANDSHAKE2;

	return 1;
}

/*
 *	Driver initialization code.
 *
 *	Ports of the first board are mapped to minor numbers 0..7,
 *	of the second to 8..15, third to 16..23 and fourth to 24..31.
 *
 *	No autoprobing is used, because it's very difficult to probe for
 *	MOXA cards (they are accessed as if it were normal memory).
 *	Card addresses are taken from the moxa_addr array.
 */

void
mx_init(void)
{
	struct moxa_port *info;
	int card;
	ulg addr, raddr;
	int port_num;
	int okay;

	printk("MOXA C-218 Serial Board Driver 0.98 (c) 1995 -- 1997 Martin Mares\n");

	/* Initialize the tty_driver structures */
    
	memset(&mx_serial_driver, 0, sizeof(struct tty_driver));
	mx_serial_driver.magic = TTY_DRIVER_MAGIC;
	mx_serial_driver.name = "ttyM";
	mx_serial_driver.major = MOXA_MAJOR;
	mx_serial_driver.minor_start = 0;
	mx_serial_driver.type = TTY_DRIVER_TYPE_SERIAL;
	mx_serial_driver.subtype = SERIAL_TYPE_NORMAL;
	mx_serial_driver.init_termios = tty_std_termios;
	mx_serial_driver.init_termios.c_cflag =
		B9600 | CS8 | CREAD | HUPCL | CLOCAL;
#ifdef MOXA_LOOPBACK_ONLY
	mx_serial_driver.init_termios.c_lflag &=
		~(ECHO | ECHOKE | ECHOK | ECHONL | ECHOKE | ECHOCTL | ECHOPRT);
#endif
	mx_serial_driver.flags = TTY_DRIVER_REAL_RAW;
	mx_serial_driver.refcount = &serial_refcount;
	mx_serial_driver.table = serial_table;
	mx_serial_driver.termios = serial_termios;
	mx_serial_driver.termios_locked = serial_termios_locked;
	mx_serial_driver.open = mx_open;
	mx_serial_driver.close = mx_close;
	mx_serial_driver.write = mx_write;
	mx_serial_driver.put_char = mx_put_char;
	mx_serial_driver.flush_chars = mx_flush_chars;
	mx_serial_driver.write_room = mx_write_room;
	mx_serial_driver.chars_in_buffer = mx_chars_in_buffer;
	mx_serial_driver.flush_buffer = mx_flush_buffer;
	mx_serial_driver.ioctl = mx_ioctl;
	mx_serial_driver.set_termios = mx_set_termios;
	mx_serial_driver.stop = mx_stop;
	mx_serial_driver.start = mx_start;
	mx_serial_driver.hangup = mx_hangup;

	/*
	 * The callout device is just like normal device except for
	 * major number and the subtype code.
	 */
	  
	mx_callout_driver = mx_serial_driver;
	mx_callout_driver.name = "cum";
	mx_callout_driver.major = MOXAAUX_MAJOR;
	mx_callout_driver.subtype = SERIAL_TYPE_CALLOUT;

	/*
	 * Initialize the interrupt handler.
	 */

	if (request_irq(MOXA_IRQ, mx_interrupt, 0, "moxa", NULL))
		panic("Cannot get IRQ\n");
	disable_irq(MOXA_IRQ);

	/*
	 * Initialize the hardware (search for cards using supplied address table).
	 */

	printk("Initializing MOXA cards:");
	port_num=0;
	okay=0;
	for(card=0;card<4;card++)
		if ((addr = moxa_addr[card])) {
			okay++;
			printk(" %X",addr);
			moxa_addr[card] = raddr = (ulg) phys_to_virt(addr);
			if (mx_card_init((byte *) raddr)) {
				int i;
				mx_events[card] = (struct moxa_event_ctrl *) raddr;
				for(i=0;i<8;i++) {
					info=&mx_port[port_num];
					info->magic = MOXA_MAGIC;
					info->ctrl = (struct moxa_chan_ctrl *) (raddr+0x400+0x60*i);
					info->page_port = (byte *) (raddr + 0x1ff0);
					info->mem_window = (byte *) (raddr + 0x2000);
					info->read_buffer = (0x90000 + 0xc000 * i);
					info->write_buffer = info->read_buffer + 0x8000;
					info->line = port_num;
					info->flags = 0;
					info->tty = 0;
					info->event = 0;
					info->count = 0;
					info->type = PORT_CIRRUS;
					info->blocked_open = 0;
					info->tqueue.routine = do_softint;
					info->tqueue.data = info;
					info->callout_termios = mx_callout_driver.init_termios;
					info->normal_termios = mx_serial_driver.init_termios;
					info->open_wait = 0;
					info->close_wait = 0;
					port_num++;
				}
			}
		}
	if (!okay) printk(" none");
	printk(".\n");
	num_channels = port_num;
	mx_serial_driver.num = mx_callout_driver.num = port_num;

	/*
	 * Register the tty's.
	 */

	if (tty_register_driver(&mx_serial_driver))
		panic("Couldn't register MOXA serial driver\n");
	if (tty_register_driver(&mx_callout_driver))
		panic("Couldn't register MOXA callout driver\n");

	/*
	 * Initialize timer mechanism and enable the interrupts.
	 */

	init_timer(&mx_timer);
	mx_timer.expires = jiffies + MOXA_TIMER_PERIOD;
	mx_timer.function = mx_timer_int;
	add_timer(&mx_timer);
	enable_irq(MOXA_IRQ);

	/*
	 * Switch local hardware loopback when requested by debugging flags.
	 */

#ifdef MOXA_LOOPBACK_ONLY
	{
		int i;
		for (i=0; i<num_channels; i++) {
			info=&mx_port[i];
			mx_cmd(info, MOXA_CMD_LOOPBACK_ON);
		}
	}
	printk("moxa: Internal hardware loopback active.\n");
#endif
}

/*
 *	Driver cleanup code. We need cleanup only when we're installed
 *	as a kernel module. If working inside the kernel, no cleanup exists.
 */

#ifdef MODULE

static void
mx_cleanup(void)
{
	if (tmp_buf)
		free_page((unsigned long) tmp_buf);
	del_timer(&mx_timer);
	tty_unregister_driver(&mx_serial_driver);
	tty_unregister_driver(&mx_callout_driver);
	free_irq(MOXA_IRQ, NULL);
}

#endif

/*
 *	Module initialization and cleanup
 */

#ifdef MODULE

int
init_module(void)
{
	mx_init();
	return 0;
}

void
cleanup_module(void)
{
	DBPRINT(("moxa: cleanup_module called.\n"));

	mx_cleanup();
}

#endif
